;@GOTO TRANSLATE

; This ExD installs own Handler for Exception 13 (caused in V86) with two
; functions:
; 1) To provide emulation of instruction MOV from/to Control/Debug register
; 2) To provide Uninstall feature

; Emulation is done by checking OpCode which caused exception 13: if OpCode
; was XX YY .. AA ZZ 0F 20/21/22/23 RR then prefixes (XX .. ZZ) and 0F byte are
; skipped.
; If the next byte is not 20..23 then control is passed to the next handler
; otherwise from byte 20..23 is determined if instruction is from/to
; from byte RR is determined control/debug register and integer register.
; V86 IP is updated to point to instruction following MOV from/to CR/DR.
; Side effect: This routine can handle instructions longer than 15 bytes.

; Uninstall is done by instruction HLT(OpCode 0F4H), when EBP is 99999999H.
; When Uninstall fails Carry flag will be set otherwise will be clear.
; HLT is (as MOV from/to) exception13-causing instruction.


; Install EmuXRX, then run Test.com.
; It will display "real" values of CR/DR and uninstalls EmuXRX.
; Run Test once again and you'll see only zeros as CR/DR - (EmuXRX was 
; uninstalled).


.586P
.MODEL FLAT

 INCLUDE EXD.INC

.CODE

TotalStart:                                   ;start of FLAT image for _HeapFree


        MySign   EQU   99999999H              ;for Uninstall
        _HLT     EQU   0F4H


       CheckAPI:
        CMP  [EBP].Client_EBP, MySign         ;is in V86 EBP "my" value
        JNE  NotMe13                          ;no -> ignore
 
        ;uninstall
	OR    BYTE PTR [EBP].Client_Eflags, 1 ;set V86 Carry flag

	LEA   ESI,MainInt13
	MOV   EAX,13
	VMMCall UnHook_V86_Fault              ;try to unhook
        JC    Quit                            ;failed -> return Carry set

        INC   [EBP].Client_EIP                ;V86 IP to the following
                                              ;instruction 

	PUSH  0 OFFSET TotalStart             ;parameters for _HeapFree

        ;copy "freemem" routine to another place (DOS stack)
        ;because it's risky to free space where I'm running

	MOVZX EDI,[EBP].Client_SS
	LEA   ESI,QuitCode
	SHL   EDI,4
	MOV   ECX,QuitCodeL
	ADD   EDI,[EBP].Client_ESP            ;EDI=LinearAddress(SS:SP)
	CLD
	SUB   EDI,ECX
	AND   BYTE PTR [EBP].Client_Eflags,NOT 1  ;success -> clear Carry flag
	PUSH  EDI
	REP   MOVSB                           ;copy to DOS stack

	RET   ;goto DOS stack

        ;this "freemem" routine is placed and executed in DOS stack

      QuitCode:
	VMMCall _HeapFree
	ADD   ESP,8                           ;two parameters of _HeapFree   
      Quit:
	POPAD
	POPFD
      UnInstFailed:
	RET
      QuitCodeL EQU $-QuitCode

;-------------------------------------------------------------------------------

        ;Following jumps are required  ("handler header") by Install_???_Hook
        ;(trace Install_???_Hook in WinICE) to get PrevHook13 address

	JMP	MainINT13             ;EB 06  (can be EB 0A too)

      Chain13:
	JMP	PrevHook13            ;FF 25 AdrOfPrev
                                      ;Hook will fill [AdrOfPrev] 

      MainINT13:                      ;handler begins here!!!

      ;in EBX is (as always) handle of Virtual Machine which caused exception 13
      ;in EBP is (as always) pointer to Client_Reg_Struc

	PUSHFD
	PUSHAD

	CLD
	MOVZX	ESI,[EBP].Client_CS
        SHL     ESI,4                 ;it's V86 fault - so calculate
	ADD	ESI,[EBP].Client_EIP  ;linear address by standard method
        MOV     EBX,ESI               

        CMP     BYTE PTR [ESI],_HLT   ;was it HLT (API?)
        JE      CheckAPI

        ;-------skip prefixes
       NextPref:
	LODSB
	LEA	EDI,PrefixTable       ;contains prefixes(without LOCK) + 0FH
	MOV	ECX,PrefixTableL
	REPNE	SCASB
	JE	NextPref

	MOV     DL, AL
        AND     AL, NOT 3             ;23,22,21,20 -> 20
        CMP     AL, 20H
        JNE     NotMe13               ;it's not MOV -> ignore

        ;from "3rd" byte of instruction extract XRX number and R32 number
        PUSH    EBX
	LODSB
	MOV	CL,7  
	MOV	EBX,EAX
	XCHG	EAX,ECX
	AND	CL,111B
	SUB	AL,CL
	CBW
	AND	EBX,111000B    
	CWDE
        POP     EDI


	CMP	 DL,20H
	JE	GiveXRX
	CMP	 DL,21H
	JE	GiveXRX

	CMP	 DL,22H
	JE	WritXRX
	CMP	 DL,23H
	JNE	NotMe13

        ;[EBP+EAX*4] = position of integer register in Client_Reg_struc
        ;except ESP

      WritXRX:
        MOV     MOV_XRX_EAX,BL        ;reconstruct MOV to XRX
        MOV     FuncWri,DL 

	MOV	EAX,[EBP+EAX*4]       ;integer register from Client_Reg_struc
        JECXZ   $+2                   ;avoid pipe effect
                    DB  0FH           ;this is
        FuncWri     DB  00H           ;reconstructed
        MOV_XRX_EAX DB  00H           ;MOV ?R?, EAX
        JMP     Quit2

     GiveXRX:
        INC     BL                    ;reconstruct MOV from XRX
        MOV     FuncRea,DL 
        MOV     MOV_ECX_XRX,BL 

        JECXZ   $+2                   ;avoid pipe effect
                    DB  0FH           ;this is        
        FuncRea     DB  00H           ;reconstructed
        MOV_ECX_XRX DB  00H           ;MOV ECX, ?R?
	MOV	[EBP+EAX*4],ECX       ;integer register from Client_Reg_struc

     Quit2:
        SUB     ESI,EDI               ;EDI = original IP
	ADD	[EBP].Client_EIP,ESI  ;V86 IP to the next instruction

        JMP     Quit

     NotMe13:                         ;it was not MOV from/to CR/DR or "my" HLT
	CMP	PrevHook13,0          ;no previous hook?
	JE	Quit                  ;yes -> quit RET
	POPAD
	POPFD
	JMP	Chain13               ;no -> chain

PrefixTable	DB 26H,2EH,36H,3EH,64H,65H,66H,67H,0F2H,0F3H,  0FH
PrefixTableL	EQU $-PrefixTable


Start:
	LEA     ESI,MainINT13
	MOV     EAX,13
	VMMCall Hook_V86_Fault
        ;returns CF according to result so there's no need to set/clear CF
        ;explicitly
        RET

.DATA
	PrevHook13  DD ?

.STACK	4

END Start

:TRANSLATE
@ECHO OFF
REM   CALL .BAT   PathToExDLoad PathToExDLoad ThisFolder PathToInclude
CALL  ..\..\ExDLoad\ExDLdr.bat ..\..\ExDLoad\ .\ ..\..\Include
TASM  /M /T /Q /I..\..\Include EmuXRX.BAT
DLINK -SExDLdr EmuXRX >NUL
DEL            EmuXRX.obj
DEL            ExDLdr.exe